/*
	$Id: RMAExpress.java 301 2007-05-24 14:05:49Z jari $

	Copyright (C) 2006, 2007 Jari Hkkinen
	
	This file is part of BASEPlugIns - Non-core plug-ins for BASE,
	BioArray Software Environment.
	BASEPlugIns is available at http://baseplugins.thep.lu.se/
	BASE is available at http://base.thep.lu.se/

	BASEPlugIns is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	BASEPlugIns is distributed in the hope that it will be useful, but
	WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330,
	Boston, MA  02111-1307, USA.
*/

package se.lu.thep.affymetrix;

import net.sf.basedb.core.Affymetrix;
import net.sf.basedb.core.ArrayDesign;
import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.BioAssay;
import net.sf.basedb.core.BioAssaySet;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Experiment;
import net.sf.basedb.core.File;
import net.sf.basedb.core.Include;
import net.sf.basedb.core.InvalidDataException;
import net.sf.basedb.core.InvalidUseOfNullException;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemParameterType;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.Job;
import net.sf.basedb.core.PluginDefinition;
import net.sf.basedb.core.PluginParameter;
import net.sf.basedb.core.PositionBatcher;
import net.sf.basedb.core.ProgressReporter;
import net.sf.basedb.core.RawBioAssay;
import net.sf.basedb.core.Reporter;
import net.sf.basedb.core.ReporterBatcher;
import net.sf.basedb.core.RequestInformation;
import net.sf.basedb.core.SpotBatcher;
import net.sf.basedb.core.StringParameterType;
import net.sf.basedb.core.Transformation;

import net.sf.basedb.core.data.ReporterData;

import net.sf.basedb.core.plugin.About;
import net.sf.basedb.core.plugin.AboutImpl;
import net.sf.basedb.core.plugin.AbstractPlugin;
import net.sf.basedb.core.plugin.GuiContext;
import net.sf.basedb.core.plugin.InteractivePlugin;
import net.sf.basedb.core.plugin.Plugin;
import net.sf.basedb.core.plugin.Request;
import net.sf.basedb.core.plugin.Response;

import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Orders;

import net.sf.basedb.util.parser.FlatFileParser;

import java.io.IOException;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import java.util.regex.Pattern;

/**
	 This plugin provides means to calculate the fold changes for a root bioassayset 
	 using RMAExpress. It needs the following parameters:

	 <pre>
	 String              : name; The name of the new bioassayset.
	 Experiment          : experiment; The experiment we are working on.
	 List&gt;RawBioAssay&lt;   : rawBioAssays; The raw bioassays to create bioassays for
	                       (must be part of the experiment).
	 </pre>

	 @author Jari
	 @version MAKESUBSTOFVERSIONNUMBER
	 @base.modified $Date: 2007-05-24 16:05:49 +0200 (Thu, 24 May 2007) $
*/

public class RMAExpress
	extends AbstractPlugin
	implements InteractivePlugin 
{

	public RMAExpress() {}

	@SuppressWarnings("unchecked")
	public void configure(GuiContext context, Request request, Response response)
	{
		try {
			if (request.getCommand().equals(Request.COMMAND_CONFIGURE_JOB)) {
				List<Throwable> errors = validateRequestParameters
					(getConfigureJobParameters().getParameters(), request);
				if (errors != null) {
					response.setError(errors.size() +
														" invalid parameter(s) were found in the request",
														errors);
					return;
				}
				Experiment experiment =
					(Experiment)request.getParameterValue("experiment");
				DbControl dc = sc.newDbControl();
				int numCels=0;
				try {
					experiment = Experiment.getById(dc, experiment.getId());
					if (!experiment.getRawDataType().isAffymetrix())
						throw new BaseException("Raw data type '" +
																		experiment.getRawDataType() +
																		"' is not supoorted by this plug-in.");
					List<RawBioAssay> rawBioAssays =
						(List<RawBioAssay>)request.getParameterValues("rawBioAssays");
					for (RawBioAssay rba : rawBioAssays)
						if (!experiment.isUsing(rba))
							throw new InvalidDataException
								("Raw bioassay '" + rba.getName() +
								 "' is not part of the experiment '" + experiment.getName()+"'");
					numCels=rawBioAssays.size();
				}
				finally {
					if (dc != null)
						dc.close();
				}
				storeValue(job, request, nameParameter);
				storeValue(job, request, experimentParameter);
				storeValues(job, request, rawBioAssaysParameter);
				Job.ExecutionTime execTime = Job.ExecutionTime.SHORT;
				if (numCels>10 && numCels<30)
					execTime = Job.ExecutionTime.MEDIUM;
				else if (numCels>=30)
					execTime = Job.ExecutionTime.LONG;
				response.setDone("Job configuration complete", execTime);
			}
		}
		catch (Throwable e) {
			response.setError(e.getMessage(), Arrays.asList(e));
		}
	}


	private void download(File file)
		throws IOException
	{
		java.io.FileOutputStream os = new java.io.FileOutputStream
			(new java.io.File(getExecDirectory(),file.getName()));
		file.download(os,0);
	}


	public About getAbout() { return about; }


	private RequestInformation getConfigureJobParameters()
	{
		if (configureJob == null) {
			List<PluginParameter<?>> parameters = new ArrayList<PluginParameter<?>>();
			
			experimentType = new ItemParameterType<Experiment>(Experiment.class, null,
																												 true, 1, null);
			experimentParameter = new PluginParameter<Experiment>(
				"experiment", "Experiment", "The experiment we are working with",
				experimentType);
			parameters.add(experimentParameter);
			parameters.add(nameParameter);
			DbControl dc = sc.newDbControl();
			try {
				int experimentId = sc.getCurrentContext(Item.EXPERIMENT).getId();
				Experiment experiment = Experiment.getById(dc, experimentId);
				ItemQuery<RawBioAssay> rawBioAssayQuery = experiment.getRawBioAssays();
				rawBioAssayQuery.include(Include.MINE, Include.SHARED, Include.OTHERS,
																 Include.IN_PROJECT);
				rawBioAssayQuery.order(Orders.asc(Hql.property("name")));
				List<RawBioAssay> rawBioAssays = new ArrayList<RawBioAssay>(
																											rawBioAssayQuery.list(dc));
				rawBioAssaysType = new ItemParameterType<RawBioAssay>(RawBioAssay.class,
																															null, true, 0,
																															rawBioAssays);
				rawBioAssaysParameter = new PluginParameter<RawBioAssay>(
					"rawBioAssays", "Raw bioassays",
					"Select the raw bioassays to use when creating the root bioassay set",
					rawBioAssaysType);
				parameters.add(rawBioAssaysParameter);
			}
			finally {
				if (dc != null)
					dc.close();
			}
			configureJob = new RequestInformation(
				Request.COMMAND_CONFIGURE_JOB, "Intensity calculation options",
				"Select which raw bioassays to use and a formula for the calculations",
				parameters);
		}
		return configureJob;
	}


	private java.io.File getExecDirectory()
	{
		if (execDirectory == null) {
			execDirectory = new java.io.File
				(System.getProperty("java.io.tmpdir") + java.io.File.separator +
				 about.getName() + java.io.File.separator + job.getId());
			execDirectory.mkdirs();
		}
		return execDirectory;
	}


	public Set<GuiContext> getGuiContexts() { return guiContexts; }


	public Plugin.MainType getMainType() { return Plugin.MainType.INTENSITY; }


	public RequestInformation getRequestInformation(GuiContext context,
																									String command)
		throws BaseException
	{
		if (command.equals(Request.COMMAND_CONFIGURE_JOB))
			return getConfigureJobParameters();
		return null;
	}


	public String getRMAExpressConsolePath()
	{
		if (RMAExpressConsolePath == null) {
			DbControl dc = sc.newDbControl();
			String jarpath=PluginDefinition.getByClassName(dc,
																"se.lu.thep.affymetrix.RMAExpress").getJarPath();
			int pos = jarpath.length();
			while ((pos>0) && (jarpath.charAt(pos-1)!=java.io.File.separatorChar))
				--pos;
			RMAExpressConsolePath=jarpath.substring(0,pos);
			// Hard coded file name and the binary must reside in the same
			// directory as the plug-in jar. The idea behind this is that
			// only the BASE administrator is allowed to change files in the
			// BASE application direcotry structure.
			RMAExpressConsolePath+="RMAExpressConsole";
		}
		return RMAExpressConsolePath;
	}


	public String isInContext(GuiContext context, Object item)
	{
		DbControl dc = sc.newDbControl();
		try {
			Experiment e=Experiment.getById
				(dc, sc.getCurrentContext(Item.EXPERIMENT).getId());
			if (!e.getRawDataType().isAffymetrix())
				return ("Raw data type '" + e.getRawDataType() +
								"' is not supoorted by this plug-in.");
		}
		catch (Throwable e) {
		}
		finally {
			if (dc != null)
				dc.close();
		}
		return null;
	}


	public boolean requiresConfiguration() { return false; }


	@SuppressWarnings("unchecked")
	public void run(Request request, Response response,	ProgressReporter progress)
	{
		DbControl dc=sc.newDbControl();
		try {
			String name=(String)job.getValue("name");
			if (name==null)
				name="New root bioassay set";
			Experiment experiment=(Experiment)job.getValue("experiment");
			if (experiment==null)
				throw new InvalidUseOfNullException("experiment");
			experiment=Experiment.getById(dc, experiment.getId());
			List<RawBioAssay> sources=(List<RawBioAssay>)job.getValues("rawBioAssays");
			if (sources==null || sources.size()==0)
				throw new InvalidUseOfNullException("sources");
			for (RawBioAssay rba : sources)
				dc.reattachItem(rba);
			runRMAExpress(sources,progress);
			storeResult(dc,experiment,sources,name,progress);
			dc.commit();
			response.setDone("Plug-in ended successfully");
		}
		catch (Throwable e) {
			response.setError(e.getMessage(), Arrays.asList(e));
		}
		finally {
			// clean up
			java.io.File[] files=getExecDirectory().listFiles();
			for (java.io.File f : files)
				f.delete();
			getExecDirectory().delete();
			if (dc != null)
				dc.close();
		}
	}


 	private void runRMAExpress(List<RawBioAssay> sources,
														 ProgressReporter progress)
		throws InterruptedException, IOException
	{
		// Are filenames used here unique? There are many possible
		// clashes, maybe the implementation should use system generated
		// unique filenames?
		String filelistname="inputfiles.txt";
		java.io.PrintWriter filelist =
			new java.io.PrintWriter(new java.io.File(getExecDirectory(),filelistname));
		File cdf=Affymetrix.getCdfFile(sources.get(0).getArrayDesign());
		progress.display(percentDone,"Downloading CDF file '" +cdf.getName()+ "'.");
		filelist.println(cdf.getName());
		download(cdf);
		for (RawBioAssay rba : sources) {
			File cel=Affymetrix.getCelFile(rba);
			progress.display(percentDone,"Downloading CEL file '"+cel.getName()+"'.");
			filelist.println(cel.getName());
			download(cel);
		}
		filelist.close();
		String optionname="option.txt";
		String resultname="./rma.txt";
		java.io.PrintWriter option =
			new java.io.PrintWriter(new java.io.File(getExecDirectory(), optionname));
		option.println("1\n"+resultname);
		option.close();
		percentDone+=1;
		progress.display(percentDone,"Running RMAExpress.");
		Process p = Runtime.getRuntime().exec(getRMAExpressConsolePath() + " " +
																					filelistname + " " + optionname, null,
																					getExecDirectory());
		p.waitFor();
		int status=p.exitValue();
		if (status>0)
			throw new IOException("Unexpected exit of RMAExpress sub-process. " +
														"Return value: " + status);
	}


	private void storeResult(DbControl dc, Experiment experiment,
													 List<RawBioAssay> sources, String name,
													 ProgressReporter progress)
	{
		try {
			FlatFileParser ffp=new FlatFileParser();
			ffp.setInputStream(new java.io.FileInputStream
												 (getExecDirectory()+java.io.File.separator+"rma.txt"),
												 null);
			ffp.setDataHeaderRegexp(Pattern.compile("Probesets\\t.*"));
			ffp.setDataSplitterRegexp(Pattern.compile("\\t"));
			ffp.setMinDataColumns(sources.size()+1);
			ffp.setMaxDataColumns(sources.size()+1);
			ffp.parseHeaders();

			Transformation t=experiment.newTransformation(Job.getById(dc,job.getId()),
																										sources);
			t.setName("RMAExpress");
			BioAssaySet root = t.newProduct("new", "new", false);
			root.setName(name);
			dc.saveItem(t);
			dc.saveItem(root);

			HashMap<Integer,Short> cubecolumn=new HashMap<Integer,Short>();
			for (RawBioAssay rba : sources) {
				BioAssay ba = root.newRootBioAssay(Collections.singleton(rba));
				ba.setName(rba.getName());
				dc.saveItem(ba);
				cubecolumn.put
					(ffp.getColumnHeaderIndex(Affymetrix.getCelFile(rba).getName()),
					 ba.getDataCubeColumnNo());
			}

			int position=1;	// Use 1 for consistency with BASE usage if position
			ReporterBatcher rbatcher = ReporterBatcher.getNew(dc);
			PositionBatcher pbatcher = root.getPositionBatcher();
			SpotBatcher sbatcher = root.getSpotBatcher();
			// The progress should be inside the loop, but for now we do not
			// know how many data entries there is to store (currently the
			// number of "spots" is not stored in the database for
			// Affymetrix).
			percentDone=95;
			progress.display(percentDone,"Importing result.");
			while (ffp.hasMoreData()) {
				FlatFileParser.Data data = ffp.nextData();
				for (Map.Entry<Integer,Short> entry : cubecolumn.entrySet())
					sbatcher.insert(entry.getValue().shortValue(), position,
													Float.parseFloat(data.get(entry.getKey())));
				// The reporter id is assumed to be the first column
				pbatcher.insert(position,rbatcher.getByExternalId(data.get(0)));
				++position;
			}
			rbatcher.close();
			pbatcher.close();
			sbatcher.close();
		}
		catch(Throwable e) {
			throw new BaseException("Unable to import root bioassay.",e);
		}
	}


	public boolean supportsConfigurations() { return false; }


	private static final About about = new AboutImpl
		("RMAExpress plug-in", "Computes gene expression summary values for " +
		 "Affymetrix Genechip data using the Robust Multichip Average " +
		 "expression summary. Results are expressed in log2 scale.",
		 "Version: MAKESUBSTOFVERSIONNUMBER $Revision: 301 $ " +
		 "$Date: 2007-05-24 16:05:49 +0200 (Thu, 24 May 2007) $",
		 "2006, 2007 Jari Hakkinen, " +
		 "Department of Theoretical Physics, Lund University",
		 null, null,
		 "http://baseplugins.thep.lu.se/wiki/se.lu.thep.affymetrix" );
	private RequestInformation configureJob = null;
	private RequestInformation configurePlugin = null;
	private ItemParameterType<Experiment> experimentType;
	/**
		 This is the directory where the plug-in will generate its output.
	*/
	private java.io.File execDirectory;
	private PluginParameter<Experiment> experimentParameter;
	private static final Set<GuiContext> guiContexts = Collections.singleton
		(new GuiContext(Item.BIOASSAYSET,GuiContext.Type.LIST));
	private static final StringParameterType nameType = 
		new StringParameterType(BioAssaySet.MAX_NAME_LENGTH, "New bioassayset",
														true);
	private static final PluginParameter<String> nameParameter =
		new PluginParameter<String>("name", "Bioassay set name",
																"The name of the root bioassayset", nameType);
	private int percentDone=0;
	private ItemParameterType<RawBioAssay> rawBioAssaysType;
	private PluginParameter<RawBioAssay> rawBioAssaysParameter;
	private String RMAExpressConsolePath;
}
